import bpy
import bpy.utils.previews
import random
import os
from datetime import datetime
from bpy.app.handlers import persistent

_icons = None
cached_random_numbers = {}

def sanitize_path(path):
    invalid_chars = '<>:"/\\|?*'
    for char in invalid_chars:
        path = path.replace(char, '_')
    path = path.replace("\\", "/")
    return path[:255] if len(path) > 255 else path

def format_time(current_time, settings):
    date_parts = []
    time_parts = []
    if settings.use_year: date_parts.append(current_time[:4])
    if settings.use_month: date_parts.append(current_time[5:7])
    if settings.use_day: date_parts.append(current_time[8:10])
    if settings.use_hour: time_parts.append(current_time[11:13])
    if settings.use_minute: time_parts.append(current_time[14:16])
    if settings.use_second: time_parts.append(current_time[17:19])
    date_string = "-".join(date_parts)
    time_string = ".".join(time_parts)
    return f"({date_string} {time_string})_" if date_string or time_string else ""

def generate_new_path(settings, filename, current_time, random_number, current_frame, sequence_number, sequence_padding, i):
    path_parts = []
    if settings.use_suffix_rule and settings.custom_prefix:
        path_parts.append(settings.custom_prefix.rstrip('_') + "_")
    if settings.use_filename_rule:
        path_parts.append(filename + "_")
    if settings.use_time_rule:
        path_parts.append(format_time(current_time, settings))
    if settings.use_random_rule and random_number is not None:
        path_parts.append(f"{random_number}_")
    if settings.use_frame_rule:
        path_parts.append(f"f{current_frame}_")
    if settings.use_sequence_rule:
        path_parts.append(f"{str(sequence_number).zfill(sequence_padding)}_")
    else:
        path_parts.append(f"{i + 1:02d}_")
    new_path = "".join(path_parts).rstrip("_")
    if settings.use_suffix_rule and settings.custom_suffix:
        new_path += f"_{settings.custom_suffix.lstrip('_')}"
    return sanitize_path(new_path)

def rename_output_file_paths(node, settings, current_frame=None):
    try:
        if not node or not hasattr(node, 'file_slots'):
            return
        blend_filepath = bpy.data.filepath or "Unsaved_File"
        filename = os.path.splitext(os.path.basename(blend_filepath))[0]
        current_time = settings.cached_current_time or datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
        sequence_number = max(settings.sequence_start, 1)
        sequence_padding = max(settings.sequence_padding, 1)
        current_frame = current_frame or bpy.context.scene.frame_current

        for i, file_slot in enumerate(node.file_slots):
            random_number = None
            key = (node.as_pointer(), i)
            if settings.use_random_rule:
                if settings.auto_random:
                    if key not in cached_random_numbers:
                        random_number = random.randint(10001, 99999)
                        cached_random_numbers[key] = random_number
                    else:
                        random_number = cached_random_numbers[key]
                else:
                    if key in cached_random_numbers:
                        random_number = cached_random_numbers[key]
                    else:
                        random_number = random.randint(10001, 99999)
                        cached_random_numbers[key] = random_number
            new_path = generate_new_path(settings, filename, current_time, random_number,
                                         current_frame, sequence_number, sequence_padding, i)
            file_slot.path = new_path
            if settings.use_sequence_rule:
                sequence_number += 1
    except Exception as e:
        print(f"Error in rename_output_file_paths: {e}")

def sync_node_selection(settings, node_tree):
    try:
        existing_selections = {}
        for item in settings.node_selection:
            node = node_tree.nodes.get(item.node_name)
            if node:
                existing_selections[node.as_pointer()] = (item.node_name, item.select)
        settings.node_selection.clear()
        for node in node_tree.nodes:
            if node.type == 'OUTPUT_FILE':
                item = settings.node_selection.add()
                item.node_name = node.name
                _, select_state = existing_selections.get(node.as_pointer(), (node.name, False))
                item.select = select_state
    except Exception as e:
        print(f"Error in sync_node_selection: {e}")

def naming_rules_update_callback(self, context):
    update_naming_rules(self, context)

def update_naming_rules(settings, context):
    if getattr(settings, 'updating', False):
        return
    settings.updating = True
    try:
        if not settings.use_frame_rule:
            pass
        if not settings.use_random_rule and settings.auto_random:
            settings.auto_random = False
        if not settings.use_sequence_rule:
            if settings.sequence_start != 1:
                settings.sequence_start = 1
            if settings.sequence_padding != 3:
                settings.sequence_padding = 3
        if not settings.use_suffix_rule:
            if settings.custom_prefix:
                settings.custom_prefix = ""
            if settings.custom_suffix:
                settings.custom_suffix = ""
        if not settings.use_time_rule:
            if any([settings.use_year, settings.use_month, settings.use_day, settings.use_hour, settings.use_minute, settings.use_second]):
                settings.use_year = True
                settings.use_month = True
                settings.use_day = True
                settings.use_hour = True
                settings.use_minute = True
                settings.use_second = True
        node_tree = context.scene.node_tree
        if settings.use_multiple_nodes and node_tree:
            for item in settings.node_selection:
                if item.select:
                    node = node_tree.nodes.get(item.node_name)
                    if node and node.type == 'OUTPUT_FILE':
                        rename_output_file_paths(node, settings)
        else:
            if node_tree:
                node = node_tree.nodes.active
                if node and node.type == 'OUTPUT_FILE':
                    rename_output_file_paths(node, settings)
    except Exception as e:
        print(f"Error in update_naming_rules: {e}")
    finally:
        settings.updating = False

def update_node_selection_item_callback(self, context):
    update_naming_rules(self, context)

class NodeSelectionItem(bpy.types.PropertyGroup):
    node_name: bpy.props.StringProperty()
    select: bpy.props.BoolProperty(name="Select", default=False, update=update_node_selection_item_callback)

class NODE_OT_RenameNode(bpy.types.Operator):
    bl_idname = "node.rename_node"
    bl_label = "Rename Node"
    bl_description = "Rename the selected File Output node"
    bl_options = {'REGISTER', 'UNDO'}

    node_name: bpy.props.StringProperty()
    new_name: bpy.props.StringProperty(name="New Node Name", default="")

    def execute(self, context):
        try:
            node_tree = context.scene.node_tree
            node = node_tree.nodes.get(self.node_name)
            if node and self.new_name:
                if self.new_name in node_tree.nodes:
                    self.report({'ERROR'}, f"A node with name '{self.new_name}' already exists.")
                    return {'CANCELLED'}
                node.name = self.new_name
                settings = context.scene.naming_rules_settings
                for item in settings.node_selection:
                    if item.node_name == self.node_name:
                        item.node_name = self.new_name
                        break
                return {'FINISHED'}
            else:
                self.report({'ERROR'}, "Invalid node or name.")
                return {'CANCELLED'}
        except Exception as e:
            self.report({'ERROR'}, f"Error renaming node: {e}")
            return {'CANCELLED'}

    def invoke(self, context, event):
        self.new_name = ""
        return context.window_manager.invoke_props_dialog(self)

class NamingRulesSettings(bpy.types.PropertyGroup):
    updating: bpy.props.BoolProperty(default=False)
    is_rendering: bpy.props.BoolProperty(default=False)
    use_filename_rule: bpy.props.BoolProperty(
        name="Use Filename Rule",
        default=False,
        description="Use the blend file name as part of the output file name",
        update=naming_rules_update_callback
    )
    use_time_rule: bpy.props.BoolProperty(
        name="Use Time Rule",
        default=False,
        description="Include the current time in the output file name",
        update=naming_rules_update_callback
    )
    use_random_rule: bpy.props.BoolProperty(
        name="Use Random Rule",
        default=False,
        description="Add a random number to the output file name",
        update=naming_rules_update_callback
    )
    auto_random: bpy.props.BoolProperty(
        name="Auto",
        default=False,
        description="Automatically generate new random number during rendering",
        update=naming_rules_update_callback
    )
    use_sequence_rule: bpy.props.BoolProperty(
        name="Use Sequence Rule",
        default=False,
        description="Add a sequence number to the output file name",
        update=naming_rules_update_callback
    )
    use_frame_rule: bpy.props.BoolProperty(
        name="Use Frame Rule",
        default=False,
        description="Add the current frame number to the output file name",
        update=naming_rules_update_callback
    )
    use_suffix_rule: bpy.props.BoolProperty(
        name="Use Suffix",
        default=False,
        description="Enable custom prefix and suffix for the output file name",
        update=naming_rules_update_callback
    )
    sequence_start: bpy.props.IntProperty(
        name="Sequence Start",
        default=1,
        min=1,
        description="The starting number for the sequence rule",
        update=naming_rules_update_callback
    )
    sequence_padding: bpy.props.IntProperty(
        name="Sequence Padding",
        default=3,
        min=1,
        description="The number of digits for the sequence number (e.g., 001, 002)",
        update=naming_rules_update_callback
    )
    custom_prefix: bpy.props.StringProperty(
        name="Custom Prefix",
        default="",
        description="A custom prefix to be added before the output file name",
        update=naming_rules_update_callback
    )
    custom_suffix: bpy.props.StringProperty(
        name="Custom Suffix",
        default="",
        description="A custom suffix to be added after the output file name",
        update=naming_rules_update_callback
    )
    use_year: bpy.props.BoolProperty(
        name="Year",
        default=True,
        description="Include the current year in the output file name",
        update=naming_rules_update_callback
    )
    use_month: bpy.props.BoolProperty(
        name="Month",
        default=True,
        description="Include the current month in the output file name",
        update=naming_rules_update_callback
    )
    use_day: bpy.props.BoolProperty(
        name="Day",
        default=True,
        description="Include the current day in the output file name",
        update=naming_rules_update_callback
    )
    use_hour: bpy.props.BoolProperty(
        name="Hour",
        default=True,
        description="Include the current hour in the output file name",
        update=naming_rules_update_callback
    )
    use_minute: bpy.props.BoolProperty(
        name="Minute",
        default=True,
        description="Include the current minute in the output file name",
        update=naming_rules_update_callback
    )
    use_second: bpy.props.BoolProperty(
        name="Second",
        default=True,
        description="Include the current second in the output file name",
        update=naming_rules_update_callback
    )
    use_multiple_nodes: bpy.props.BoolProperty(
        name="Use Multiple Nodes",
        default=False,
        description="Apply naming rules to multiple File Output nodes",
        update=naming_rules_update_callback
    )
    node_selection: bpy.props.CollectionProperty(type=NodeSelectionItem)
    node_selection_index: bpy.props.IntProperty()
    cached_current_time: bpy.props.StringProperty(default="")
    last_frame: bpy.props.IntProperty(default=-1)

class NODE_PT_CustomOutputPath(bpy.types.Panel):
    bl_label = "Custom Output Path"
    bl_idname = "NODE_PT_custom_output_path"
    bl_space_type = 'NODE_EDITOR'
    bl_region_type = 'UI'
    bl_category = "Node"
    bl_context = "node"

    @classmethod
    def poll(cls, context):
        active_node = context.scene.node_tree.nodes.active
        return (context.space_data.tree_type == 'CompositorNodeTree' and
                context.scene.use_nodes and
                active_node is not None and 
                active_node.type == 'OUTPUT_FILE')

    def draw(self, context):
        layout = self.layout
        settings = context.scene.naming_rules_settings

        layout.enabled = not settings.is_rendering

        row = layout.row(align=True)
        row.prop(settings, "use_filename_rule", text="Filename", toggle=True)
        row.prop(settings, "use_time_rule", text="Time", toggle=True)
        row.prop(settings, "use_random_rule", text="Random", toggle=True)
        row.prop(settings, "use_frame_rule", text="Frame", toggle=True)

        row = layout.row(align=True)
        row.prop(settings, "use_sequence_rule", text="Sequence", toggle=True)
        row.prop(settings, "use_suffix_rule", text="Suffix", toggle=True)
        row.prop(settings, "use_multiple_nodes", text="Multiple Nodes", toggle=True)

        if settings.use_time_rule:
            time_box = layout.box()
            time_box.label(text="Time Options:")
            time_row = time_box.row(align=True)
            time_row.prop(settings, "use_year", text="Year")
            time_row.prop(settings, "use_month", text="Month")
            time_row.prop(settings, "use_day", text="Day")
            time_row = time_box.row(align=True)
            time_row.prop(settings, "use_hour", text="Hour")
            time_row.prop(settings, "use_minute", text="Minute")
            time_row.prop(settings, "use_second", text="Second")

        if settings.use_random_rule:
            random_box = layout.box()
            random_box.label(text="Random Options:")
            random_box.prop(settings, "auto_random", text="Auto", toggle=True)

        if settings.use_sequence_rule:
            seq_box = layout.box()
            seq_box.label(text="Sequence Options:")
            seq_row = seq_box.row(align=True)
            seq_row.prop(settings, "sequence_start", text="Start")
            seq_row.prop(settings, "sequence_padding", text="Padding")

        if settings.use_suffix_rule:
            suffix_box = layout.box()
            suffix_box.label(text="Suffix Options:")
            suffix_box.prop(settings, "custom_prefix", text="Custom Prefix")
            suffix_box.prop(settings, "custom_suffix", text="Custom Suffix")

        if settings.use_multiple_nodes:
            node_box = layout.box()
            node_box.label(text="Select File Output Nodes:")
            for item in settings.node_selection:
                node = context.scene.node_tree.nodes.get(item.node_name)
                if node:
                    row = node_box.row()
                    row.prop(item, "select", text=node.name)
                    row.operator("node.rename_node", text="Rename").node_name = node.name

        blend_filepath = bpy.data.filepath
        if not blend_filepath:
            warning_row = layout.row()
            warning_row.label(text="Warning: Current file is not saved!", icon='ERROR')

@persistent
def render_init_handler(dummy):
    global cached_random_numbers
    try:
        cached_random_numbers.clear()
        for scene in bpy.data.scenes:
            settings = scene.naming_rules_settings
            settings.is_rendering = True
            settings.cached_current_time = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
            node_tree = scene.node_tree
            if settings.auto_random and settings.use_random_rule:
                if settings.use_multiple_nodes:
                    for item in settings.node_selection:
                        if item.select:
                            node = node_tree.nodes.get(item.node_name)
                            if node and node.type == 'OUTPUT_FILE':
                                for i, file_slot in enumerate(node.file_slots):
                                    key = (node.as_pointer(), i)
                                    if key not in cached_random_numbers:
                                        cached_random_numbers[key] = random.randint(10001, 99999)
                                rename_output_file_paths(node, settings)
                else:
                    node = node_tree.nodes.active
                    if node and node.type == 'OUTPUT_FILE':
                        for i, file_slot in enumerate(node.file_slots):
                            key = (node.as_pointer(), i)
                            if key not in cached_random_numbers:
                                cached_random_numbers[key] = random.randint(10001, 99999)
                        rename_output_file_paths(node, settings)
    except Exception as e:
        print(f"Error in render_init_handler: {e}")

@persistent
def render_complete_handler(scene):
    try:
        for scene in bpy.data.scenes:
            settings = scene.naming_rules_settings
            settings.is_rendering = False
        cached_random_numbers.clear()
    except Exception as e:
        print(f"Error in render_complete_handler: {e}")

@persistent
def render_cancel_handler(scene):
    try:
        for scene in bpy.data.scenes:
            settings = scene.naming_rules_settings
            settings.is_rendering = False
        cached_random_numbers.clear()
    except Exception as e:
        print(f"Error in render_cancel_handler: {e}")

@persistent
def frame_change_post_handler(scene):
    try:
        settings = scene.naming_rules_settings
        node_tree = scene.node_tree
        needs_update = False

        if settings.use_random_rule and settings.auto_random:
            needs_update = True

        if settings.use_frame_rule:
            needs_update = True

        if needs_update and settings.is_rendering:
            current_frame = scene.frame_current
            if settings.last_frame != current_frame:
                settings.last_frame = current_frame
                if settings.use_multiple_nodes:
                    for item in settings.node_selection:
                        if item.select:
                            node = node_tree.nodes.get(item.node_name)
                            if node and node.type == 'OUTPUT_FILE':
                                if settings.auto_random and settings.use_random_rule:
                                    for i, file_slot in enumerate(node.file_slots):
                                        key = (node.as_pointer(), i)
                                        cached_random_numbers[key] = random.randint(10001, 99999)
                                rename_output_file_paths(node, settings, current_frame=current_frame)
                else:
                    node = node_tree.nodes.active
                    if node and node.type == 'OUTPUT_FILE':
                        if settings.auto_random and settings.use_random_rule:
                            for i, file_slot in enumerate(node.file_slots):
                                key = (node.as_pointer(), i)
                                cached_random_numbers[key] = random.randint(10001, 99999)
                        rename_output_file_paths(node, settings, current_frame=current_frame)
    except Exception as e:
        print(f"Error in frame_change_post_handler: {e}")

@persistent
def depsgraph_update_handler(scene, depsgraph):
    try:
        for scene in bpy.data.scenes:
            settings = scene.naming_rules_settings
            if settings.use_multiple_nodes:
                node_tree = scene.node_tree
                if node_tree:
                    sync_node_selection(settings, node_tree)
    except Exception as e:
        print(f"Error in depsgraph_update_handler: {e}")

def register_handler(handler, handler_list):
    if handler not in handler_list:
        handler_list.append(handler)

def unregister_handler(handler, handler_list):
    if handler in handler_list:
        handler_list.remove(handler)

def register():
    global _icons
    _icons = bpy.utils.previews.new()
    bpy.utils.register_class(NodeSelectionItem)
    bpy.utils.register_class(NamingRulesSettings)
    bpy.utils.register_class(NODE_OT_RenameNode)
    bpy.utils.register_class(NODE_PT_CustomOutputPath)
    bpy.types.Scene.naming_rules_settings = bpy.props.PointerProperty(type=NamingRulesSettings)

    register_handler(render_init_handler, bpy.app.handlers.render_init)
    register_handler(render_complete_handler, bpy.app.handlers.render_complete)
    register_handler(render_cancel_handler, bpy.app.handlers.render_cancel)
    register_handler(frame_change_post_handler, bpy.app.handlers.frame_change_post)
    register_handler(depsgraph_update_handler, bpy.app.handlers.depsgraph_update_post)

def unregister():
    global _icons
    bpy.utils.previews.remove(_icons)
    bpy.utils.unregister_class(NodeSelectionItem)
    bpy.utils.unregister_class(NamingRulesSettings)
    bpy.utils.unregister_class(NODE_OT_RenameNode)
    bpy.utils.unregister_class(NODE_PT_CustomOutputPath)
    del bpy.types.Scene.naming_rules_settings

    unregister_handler(render_init_handler, bpy.app.handlers.render_init)
    unregister_handler(render_complete_handler, bpy.app.handlers.render_complete)
    unregister_handler(render_cancel_handler, bpy.app.handlers.render_cancel)
    unregister_handler(frame_change_post_handler, bpy.app.handlers.frame_change_post)
    unregister_handler(depsgraph_update_handler, bpy.app.handlers.depsgraph_update_post)

if __name__ == "__main__":
    register()
